@@ -13,11 +13,14 @@
 
 #include <errno.h>
 #include <openssl/crypto.h>
+#include <openssl/evp.h>
 #include <stdint.h>
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
 #include <syslog.h>
+#include <dlfcn.h>
+#include <link.h>
 
 #include <openssl/opensslconf.h>
 #ifdef OPENSSL_FIPS
@@ -27,9 +30,49 @@
 #include "fips.h"
 #include "ica_api.h"
 #include "test_vec.h"
+#include "s390_crypto.h"
+
+#if OPENSSL_VERSION_PREREQ(3, 0)
+#include <openssl/crypto.h>
+#include <openssl/provider.h>
+extern OSSL_LIB_CTX *openssl_libctx;
+extern OSSL_PROVIDER *openssl_provider;
+#endif
+
+int openssl_in_fips_mode(void)
+{
+#if !OPENSSL_VERSION_PREREQ(3, 0)
+	return FIPS_mode();
+#else
+	if (fips & ICA_FIPS_INTEGRITY)
+		return 0;
+	else
+		return 1;
+#endif
+}
+
+#ifndef PATH_MAX
+#define PATH_MAX 4096
+#endif
+
+#define HMAC_PREFIX "."
+#define HMAC_SUFFIX ".hmac"
+#define READ_BUFFER_LENGTH 16384
+
+/*
+ * The hard-coded HMAC key to be optionally provided for the library
+ * integrity test. The recommended key size for HMAC-SHA256 is 64 bytes.
+ * The known HMAC is supposed to be provided as hex string in a file
+ * .libica.so.VERSION.hmac in the same directory as the .so module.
+ */
+static const char hmackey[] =
+	"0000000000000000000000000000000000000000000000000000000000000000"
+	"0000000000000000000000000000000000000000000000000000000000000000";
 
 int fips;
 
+#define LIBICA_FIPS_CONFIG		LIBICA_CONFDIR "/libica/openssl3-fips.cnf"
+
 static int aes_ecb_kat(void);
 static int aes_cbc_kat(void);
 static int aes_cbc_cs_kat(void);
@@ -51,6 +94,7 @@ static int des3_cmac_kat(void);
 
 static int rsa_kat(void);
 
+#ifndef NO_CPACF
 #define SHA_KAT(_sha_, _ctx_)						\
 static int sha##_sha_##_kat(void) {					\
 	sha##_ctx_##_context_t ctx;					\
@@ -74,6 +118,92 @@ SHA_KAT(256, 256);
 SHA_KAT(384, 512);
 SHA_KAT(512, 512);
 #undef SHA_KAT
+#else /* Don't write any error msg to syslog when CPACF is not avail */
+#define SHA_KAT(_sha_, _ctx_)						\
+static int sha##_sha_##_kat(void) {					\
+	sha##_ctx_##_context_t ctx;					\
+	size_t i;							\
+	unsigned char out[SHA##_sha_##_HASH_LENGTH];			\
+	for (i = 0; i < SHA##_sha_##_TV_LEN; i++) {			\
+		if (ica_sha##_sha_(SHA_MSG_PART_ONLY,			\
+		    SHA##_sha_##_TV[i].msg_len, SHA##_sha_##_TV[i].msg,	\
+		    &ctx, out) || memcmp(SHA##_sha_##_TV[i].md, out,	\
+		    SHA##_sha_##_HASH_LENGTH)) {			\
+			return 1;					\
+		}							\
+	}								\
+	return 0;							\
+}
+SHA_KAT(1, );
+SHA_KAT(224, 256);
+SHA_KAT(256, 256);
+SHA_KAT(384, 512);
+SHA_KAT(512, 512);
+#undef SHA_KAT
+#endif
+
+static inline int sha3_available(void)
+{
+	sha3_224_context_t sha3_224_context;
+	unsigned char output_hash[SHA3_224_HASH_LENGTH];
+	unsigned char test_data[] = { 0x61,0x62,0x63 };
+	int rc = 0;
+
+	rc = ica_sha3_224(SHA_MSG_PART_ONLY, sizeof(test_data), test_data,
+			&sha3_224_context, output_hash);
+
+	return (rc == ENODEV ? 0 : 1);
+}
+
+#ifndef NO_CPACF
+#define SHA3_KAT(_sha_, _ctx_)						\
+static int sha3_##_sha_##_kat(void) {					\
+	sha3_##_ctx_##_context_t ctx;					\
+	size_t i;							\
+	unsigned char out[SHA3_##_sha_##_HASH_LENGTH];			\
+	if (!sha3_available()) 						\
+		return 0; 						\
+	for (i = 0; i < SHA3_##_sha_##_TV_LEN; i++) {			\
+		if (ica_sha3_##_sha_(SHA_MSG_PART_ONLY,			\
+		    SHA3_##_sha_##_TV[i].msg_len, SHA3_##_sha_##_TV[i].msg,	\
+		    &ctx, out) || memcmp(SHA3_##_sha_##_TV[i].md, out,	\
+		    SHA3_##_sha_##_HASH_LENGTH)) {			\
+			syslog(LOG_ERR, "Libica SHA-3 %d test failed.",	\
+			    _sha_);					\
+			return 1;					\
+		}							\
+	}								\
+	return 0;							\
+}
+SHA3_KAT(224, 224);
+SHA3_KAT(256, 256);
+SHA3_KAT(384, 384);
+SHA3_KAT(512, 512);
+#undef SHA3_KAT
+#else /* Don't write any error msg to syslog when CPACF is not avail */
+#define SHA3_KAT(_sha_, _ctx_)						\
+static int sha3_##_sha_##_kat(void) {					\
+	sha3_##_ctx_##_context_t ctx;					\
+	size_t i;							\
+	unsigned char out[SHA3_##_sha_##_HASH_LENGTH];			\
+	if (!sha3_available()) 						\
+		return 0; 						\
+	for (i = 0; i < SHA3_##_sha_##_TV_LEN; i++) {			\
+		if (ica_sha3_##_sha_(SHA_MSG_PART_ONLY,			\
+		    SHA3_##_sha_##_TV[i].msg_len, SHA3_##_sha_##_TV[i].msg,	\
+		    &ctx, out) || memcmp(SHA3_##_sha_##_TV[i].md, out,	\
+		    SHA3_##_sha_##_HASH_LENGTH)) {			\
+			return 1;					\
+		}							\
+	}								\
+	return 0;							\
+}
+SHA3_KAT(224, 224);
+SHA3_KAT(256, 256);
+SHA3_KAT(384, 384);
+SHA3_KAT(512, 512);
+#undef SHA3_KAT
+#endif
 
 void
 fips_init(void)
@@ -91,6 +221,7 @@ fips_init(void)
 	fclose(fd);
 
 	if (fips_flag - '0') {
+#if !OPENSSL_VERSION_PREREQ(3, 0)
 		/* Set libica into FIPS mode. */
 		fips |= ICA_FIPS_MODE;
 
@@ -99,24 +230,282 @@ fips_init(void)
 		 * be disabled. OpenSSL FIPS mode can be queried using the
 		 * FIPS_mode() function. */
 		FIPS_mode_set(1);
+#else
+		fips = 0;
+		if (!OSSL_LIB_CTX_load_config(openssl_libctx, LIBICA_FIPS_CONFIG)) {
+			syslog(LOG_ERR, "Libica failed to load openssl fips config %s\n",
+					LIBICA_FIPS_CONFIG);
+			fips |= ICA_FIPS_INTEGRITY;
+			return;
+		}
+
+		openssl_provider = OSSL_PROVIDER_load(openssl_libctx, "fips");
+		if (openssl_provider == NULL) {
+			syslog(LOG_ERR, "Libica failed to load fips provider.\n");
+			fips |= ICA_FIPS_INTEGRITY;
+			return;
+		}
+
+		if (!EVP_set_default_properties(openssl_libctx, "fips=yes")) {
+			syslog(LOG_ERR, "Libica failed to set default properties 'fips=yes'\n");
+			fips |= ICA_FIPS_INTEGRITY;
+			return;
+		}
+
+		fips |= ICA_FIPS_MODE;
+#endif
 	}
 }
 
+static char *make_hmac_path(const char *origpath)
+{
+	char *path;
+	const char *fn;
+
+	path = calloc(1, sizeof(HMAC_PREFIX) + sizeof(HMAC_SUFFIX) + strlen(origpath) + 1);
+	if (path == NULL)
+		return NULL;
+
+	fn = strrchr(origpath, '/');
+	if (fn == NULL) {
+		fn = origpath;
+	} else {
+		++fn;
+	}
+
+	strncpy(path, origpath, fn - origpath);
+	strcat(path, HMAC_PREFIX);
+	strcat(path, fn);
+	strcat(path, HMAC_SUFFIX);
+
+	return path;
+}
+
+static int compute_file_hmac(const char *path, void **buf, size_t *hmaclen)
+{
+	FILE *fp = NULL;
+	int rc = -1;
+	unsigned char rbuf[READ_BUFFER_LENGTH];
+	unsigned char *keybuf;
+	EVP_MD_CTX *mdctx = NULL;
+	EVP_PKEY *pkey = NULL;
+	size_t hlen, len;
+	long keylen;
+
+	*buf = NULL;
+	*hmaclen = 0;
+
+	keybuf = OPENSSL_hexstr2buf(hmackey, &keylen);
+	pkey = EVP_PKEY_new_mac_key(EVP_PKEY_HMAC, NULL, keybuf, (int)keylen);
+	if (!pkey)
+		goto end;
+
+	mdctx = EVP_MD_CTX_create();
+	if (!mdctx)
+		goto end;
+
+	fp = fopen(path, "r");
+	if (fp == NULL)
+		goto end;
+
+	if (EVP_DigestSignInit(mdctx, NULL, EVP_sha256(), NULL, pkey) <= 0)
+		goto end;
+
+	while ((len = fread(rbuf, 1, sizeof(rbuf), fp)) != 0) {
+		if (EVP_DigestSignUpdate(mdctx, rbuf, len) <= 0) {
+			goto end;
+		}
+	}
+
+	if (EVP_DigestSignFinal(mdctx, rbuf, &hlen) <= 0)
+		goto end;
+
+	*buf = malloc(hlen);
+	if (*buf == NULL)
+		goto end;
+
+	*hmaclen = hlen;
+
+	memcpy(*buf, rbuf, hlen);
+
+	rc = 0;
+
+end:
+
+	if (pkey != NULL)
+		EVP_PKEY_free(pkey);
+
+	OPENSSL_free(keybuf);
+	EVP_MD_CTX_destroy(mdctx);
+	if (fp)
+		fclose(fp);
+
+	return rc;
+}
+
+/**
+ * Performs the FIPS check.
+ *
+ * @return  1 if check succeeded
+ *          0 otherwise
+ */
+static int FIPSCHECK_verify(const char *path)
+{
+	int rc = 0;
+	FILE *fp;
+	unsigned char *known_hmac = NULL;
+	long hmaclen;
+	char *hmacpath, *p;
+	char *known_hmac_str = NULL;
+	size_t n, computed_hmac_len;
+	void *computed_hmac = NULL;
+
+	hmacpath = make_hmac_path(path);
+	if (hmacpath == NULL)
+		return 0;
+
+	fp = fopen(hmacpath, "r");
+	if (fp == NULL) {
+		goto end;
+	}
+
+	if (getline(&known_hmac_str, &n, fp) <= 0)
+		goto end;
+
+	if ((p = strchr(known_hmac_str, '\n')) != NULL)
+		*p = '\0';
+
+	known_hmac = OPENSSL_hexstr2buf(known_hmac_str, &hmaclen);
+
+	if (compute_file_hmac(path, &computed_hmac, &computed_hmac_len) != 0)
+		goto end;
+
+	if (memcmp(computed_hmac, known_hmac, computed_hmac_len) != 0)
+		goto end;
+
+	rc = 1;
+
+end:
+
+	free(computed_hmac);
+	free(known_hmac_str);
+	free(hmacpath);
+
+	OPENSSL_free(known_hmac);
+
+	if (fp)
+		fclose(fp);
+
+	return rc;
+}
+
+static const char msg1[] = "Libica FIPS library integrity check failed. Cannot determine library path.\n";
+static const char msg2[] = "Libica FIPS library integrity check failed. Module %s probably corrupted.\n";
+static const char msg3[] = "Libica FIPS library integrity check passed.\n";
+
+struct phdr_cb_data {
+	/* User-provided storage for library path. */
+	char *librarypath;
+	/* Length of storage provided by user. */
+	size_t length;
+	/* How many times did we find a proper library. This is used
+	 * as a sanity check. */
+	int count;
+};
+
+static int phdr_callback(struct dl_phdr_info *info, size_t size, void *data)
+{
+	int j;
+	unsigned long start, end;
+	struct phdr_cb_data *d = data;
+	unsigned long myaddr = (unsigned long)&phdr_callback;
+
+	(void)size;
+	for (j = 0; j < info->dlpi_phnum; j++) {
+		// Only consider loadable program segments
+		if (info->dlpi_phdr[j].p_type == PT_LOAD) {
+			start = info->dlpi_addr + info->dlpi_phdr[j].p_vaddr;
+			end = start + info->dlpi_phdr[j].p_memsz;
+			if (start <= myaddr && myaddr < end) {
+				if (d->librarypath[0] == 0
+					&& strlen(info->dlpi_name) < d->length) {
+					strcpy(d->librarypath, info->dlpi_name);
+				}
+				d->count++;
+			}
+		}
+	}
+	return 0;
+}
+
+/*
+ * Perform an integrity check on libica.so by calculating an HMAC from
+ * the file contents using a static HMAC key, and comparing it to a
+ * pre-calculated HMAC in a separate file. The HMAC key and HMAC file
+ * may be provided by a Distributor when building the packet.
+ */
+static void fips_lib_integrity_check(void)
+{
+	char path[PATH_MAX];
+	struct phdr_cb_data data = {
+		.librarypath = (char *)path,
+		.length = sizeof(path),
+		.count = 0
+	};
+
+	path[0] = 0;
+	dl_iterate_phdr(phdr_callback, &data);
+	if (data.count != 1) {
+		syslog(LOG_ERR, msg1);
+		fips |= ICA_FIPS_INTEGRITY;
+		return;
+	}
+
+	if (!FIPSCHECK_verify(path)) {
+		syslog(LOG_ERR, msg2, path);
+		fips |= ICA_FIPS_INTEGRITY;
+		return;
+	}
+
+	syslog(LOG_INFO, msg3);
+}
+
 void
 fips_powerup_tests(void)
 {
+#ifdef NO_CPACF
+	/* 27 out of the 28 tests return EPERM if CPACF is disabled via config.
+	 * The rsa_kat() is not affected. */
+	int num_cpacf_tests = 27;
+#endif
+	int rc;
+
 	/* Cryptographic algorithm test. */
-	if (ica_drbg_health_test(ica_drbg_generate, 256, true, ICA_DRBG_SHA512)
-	    || sha1_kat() || sha224_kat() || sha256_kat() || sha384_kat()
-	    || sha512_kat() || des3_ecb_kat() || des3_cbc_kat()
-	    || des3_cbc_cs_kat() || des3_cfb_kat() || des3_ofb_kat()
-	    || des3_ctr_kat() || des3_cmac_kat() || aes_ecb_kat()
-	    || aes_cbc_kat() || aes_cbc_cs_kat() || aes_cfb_kat()
-	    || aes_ctr_kat() || aes_ofb_kat() || aes_ccm_kat() || aes_gcm_kat()
-	    || aes_xts_kat() || aes_cmac_kat() || rsa_kat()) {
+	rc = ica_drbg_health_test(ica_drbg_generate, 256, true, ICA_DRBG_SHA512)
+	    + sha1_kat() + sha224_kat() + sha256_kat() + sha384_kat()
+	    + sha512_kat() + sha3_224_kat() + sha3_256_kat() + sha3_384_kat()
+	    + sha3_512_kat() + des3_ecb_kat() + des3_cbc_kat()
+	    + des3_cbc_cs_kat() + des3_cfb_kat() + des3_ofb_kat()
+	    + des3_ctr_kat() + des3_cmac_kat() + aes_ecb_kat()
+	    + aes_cbc_kat() + aes_cbc_cs_kat() + aes_cfb_kat()
+	    + aes_ctr_kat() + aes_ofb_kat() + aes_ccm_kat() + aes_gcm_kat()
+	    + aes_xts_kat() + aes_cmac_kat() + rsa_kat();
+#ifndef NO_CPACF
+	if (rc != 0) {
+#else
+	if (rc != 0 && rc != num_cpacf_tests * EPERM) {
+#endif
 		fips |= ICA_FIPS_CRYPTOALG;
 		return;
 	}
+
+/* ICA internal test does not link against the library. So we should
+ * skip the library integrity check in that case.
+ */
+#ifndef ICA_INTERNAL_TEST
+	/* Library integrity test */
+	fips_lib_integrity_check();
+#endif
 }
 
 static int
@@ -146,7 +535,9 @@ aes_ecb_kat(void) {
 
 _err_:
 	free(out);
+#ifndef NO_CPACF
 	syslog(LOG_ERR, "Libica AES-ECB test failed.");
+#endif
 	return 1;
 }
 
@@ -182,7 +573,9 @@ aes_cbc_kat(void) {
 
 _err_:
 	free(out);
+#ifndef NO_CPACF
 	syslog(LOG_ERR, "Libica AES-CBC test failed.");
+#endif
 	return 1;
 }
 
@@ -220,7 +613,9 @@ aes_cbc_cs_kat(void)
 
 _err_:
 	free(out);
+#ifndef NO_CPACF
 	syslog(LOG_ERR, "Libica AES-CBC-CS test failed.");
+#endif
 	return 1;
 }
 
@@ -256,7 +651,9 @@ aes_cfb_kat(void) {
 
 _err_:
 	free(out);
+#ifndef NO_CPACF
 	syslog(LOG_ERR, "Libica AES-CFB test failed.");
+#endif
 	return 1;
 }
 
@@ -292,7 +689,9 @@ aes_ofb_kat(void) {
 
 _err_:
 	free(out);
+#ifndef NO_CPACF
 	syslog(LOG_ERR, "Libica AES-OFB test failed.");
+#endif
 	return 1;
 }
 
@@ -328,7 +727,9 @@ aes_ctr_kat(void) {
 
 _err_:
 	free(out);
+#ifndef NO_CPACF
 	syslog(LOG_ERR, "Libica AES-CTR test failed.");
+#endif
 	return 1;
 }
 
@@ -369,7 +770,9 @@ aes_ccm_kat(void) {
 _err_:
 	free(ciphertext);
 	free(payload);
+#ifndef NO_CPACF
 	syslog(LOG_ERR, "Libica AES-CCM test failed.");
+#endif
 	return 1;
 }
 
@@ -443,7 +846,7 @@ aes_gcm_kat(void) {
 			goto _err_;
 
 		memset(tag, 0, AES_BLKSIZE);
-		memset(out, 0, AES_BLKSIZE);
+		memset(out, 0, tv->len);
 		memset(icb, 0, sizeof(icb));
 		memset(icb, 0, sizeof(ucb));
 		memset(subkey, 0, sizeof(subkey));
@@ -469,7 +872,9 @@ aes_gcm_kat(void) {
 _err_:
 	free(tag);
 	free(out);
+#ifndef NO_CPACF
 	syslog(LOG_ERR, "Libica AES-GCM test failed.");
+#endif
 	return 1;
 }
 
@@ -505,7 +910,9 @@ aes_xts_kat(void) {
 
 _err_:
 	free(out);
+#ifndef NO_CPACF
 	syslog(LOG_ERR, "Libica AES-XTS test failed.");
+#endif
 	return 1;
 }
 
@@ -572,7 +979,9 @@ aes_cmac_kat(void)
 
 _err_:
 	free(mac);
+#ifndef NO_CPACF
 	syslog(LOG_ERR, "Libica AES-CMAC test failed.");
+#endif
 	return 1;
 }
 
@@ -604,7 +1013,9 @@ des3_ecb_kat(void) {
 
 _err_:
 	free(out);
+#ifndef NO_CPACF
 	syslog(LOG_ERR, "Libica 3DES-ECB test failed.");
+#endif
 	return 1;
 }
 
@@ -640,7 +1051,9 @@ des3_cbc_kat(void) {
 
 _err_:
 	free(out);
+#ifndef NO_CPACF
 	syslog(LOG_ERR, "Libica 3DES-CBC test failed.");
+#endif
 	return 1;
 }
 
@@ -677,7 +1090,9 @@ des3_cbc_cs_kat(void){
 
 _err_:
 	free(out);
+#ifndef NO_CPACF
 	syslog(LOG_ERR, "Libica 3DES-CBC-CS test failed.");
+#endif
 	return 1;
 }
 
@@ -713,7 +1128,9 @@ des3_cfb_kat(void) {
 
 _err_:
 	free(out);
+#ifndef NO_CPACF
 	syslog(LOG_ERR, "Libica 3DES-CFB test failed.");
+#endif
 	return 1;
 }
 
@@ -749,7 +1166,9 @@ des3_ofb_kat(void) {
 
 _err_:
 	free(out);
+#ifndef NO_CPACF
 	syslog(LOG_ERR, "Libica 3DES-OFB test failed.");
+#endif
 	return 1;
 }
 
@@ -785,7 +1204,9 @@ des3_ctr_kat(void) {
 
 _err_:
 	free(out);
+#ifndef NO_CPACF
 	syslog(LOG_ERR, "Libica 3DES-CTR test failed.");
+#endif
 	return 1;
 }
 
@@ -852,7 +1273,9 @@ des3_cmac_kat(void)
 
 _err_:
 	free(mac);
+#ifndef NO_CPACF
 	syslog(LOG_ERR, "Libica 3DES-CMAC test failed.");
+#endif
 	return 1;
 }
 
@@ -933,5 +1356,4 @@ _err_:
 	syslog(LOG_ERR, "Libica RSA test failed.");
 	return 1;
 }
-
 #endif /* FIPS_H */
