@@ -34,18 +34,18 @@ static stats_entry_t *stats = NULL;
 volatile int stats_shm_handle = NOT_INITIALIZED;
 
 
-static inline void atomic_add(int *x, int i)
+static inline void atomic_add(uint64_t *x, uint64_t i)
 {
-	int old;
-	int new;
-	asm volatile ("	l	%0,%2\n"
-		      "0:	lr	%1,%0\n"
-		      "	ar	%1,%3\n"
-		      "	cs	%0,%1,%2\n"
-		      "	jl	0b"
-		      :"=&d" (old), "=&d"(new), "=Q"(*x)
-		      :"d"(i), "Q"(*x)
-		      :"cc", "memory");
+    uint64_t old;
+    uint64_t new;
+    asm volatile ("   lg      %0,%2\n"
+                  "0: lgr     %1,%0\n"
+                  "   agr     %1,%3\n"
+                  "   csg     %0,%1,%2\n"
+                  "   jl      0b"
+                 :"=&d" (old), "=&d"(new), "=Q"(*x)
+                 :"d"(i), "Q"(*x)
+                 :"cc", "memory");
 }
 
 
@@ -125,7 +125,7 @@ void stats_munmap(int unlink)
  * @direction - valid values are ENCRYPT and DECRYPT
  */
 
-uint32_t stats_query(stats_fields_t field, int hardware, int direction)
+uint64_t stats_query(stats_fields_t field, int hardware, int direction)
 {
 	if (stats == NULL)
 		return 0;
@@ -277,14 +277,14 @@ void stats_increment(stats_fields_t field, int hardware, int direction)
 
 	if(direction == ENCRYPT)
 		if (hardware == ALGO_HW)
-			atomic_add((int *)&stats[field].enc.hw, 1);
+			atomic_add(&stats[field].enc.hw, 1);
 		else
-			atomic_add((int *)&stats[field].enc.sw, 1);
+			atomic_add(&stats[field].enc.sw, 1);
 	else
 		if (hardware == ALGO_HW)
-			atomic_add((int *)&stats[field].dec.hw, 1);
+			atomic_add(&stats[field].dec.hw, 1);
 		else
-			atomic_add((int *)&stats[field].dec.sw, 1);
+			atomic_add(&stats[field].dec.sw, 1);
 }
 #endif
 
