@@ -280,7 +280,7 @@ static void set_switches(int msa)
 			s390_pcc(S390_CRYPTO_QUERY, mask);
 		end_sigill_section(&oldact, &oldset);
 	}
-	for (n = 0; n < (sizeof(s390_ppno_functions) /
+	for (n = 0; n < (sizeof(s390_pcc_functions) /
 			 sizeof(s390_supported_function_t)); n++)
 		if (S390_CRYPTO_TEST_MASK(mask, s390_pcc_functions[n].hw_fc))
 			*s390_pcc_functions[n].enabled = 1;
@@ -449,13 +449,13 @@ void s390_crypto_switches_init(void)
  * The last field represents the property flags indicating key lengths
  */
 libica_func_list_element_int icaList[] = {
- {SHA1,   KIMD, SHA_1  , ICA_FLAG_SW, 0},
- {SHA224, KIMD, SHA_256, ICA_FLAG_SW, 0},
- {SHA256, KIMD, SHA_256, ICA_FLAG_SW, 0},
- {SHA384, KIMD, SHA_512, ICA_FLAG_SW, 0},
- {SHA512, KIMD, SHA_512, ICA_FLAG_SW, 0},
- {SHA512_224, KIMD, SHA_512_224, ICA_FLAG_SW, 0},
- {SHA512_256, KIMD, SHA_512_256, ICA_FLAG_SW, 0},
+ {SHA1,   KIMD, SHA_1  , 0, 0},
+ {SHA224, KIMD, SHA_256, 0, 0},
+ {SHA256, KIMD, SHA_256, 0, 0},
+ {SHA384, KIMD, SHA_512, 0, 0},
+ {SHA512, KIMD, SHA_512, 0, 0},
+ {SHA512_224, KIMD, SHA_512_224, 0, 0},
+ {SHA512_256, KIMD, SHA_512_256, 0, 0},
  {SHA3_224, KIMD, SHA_3_224, 0, 0},
  {SHA3_256, KIMD, SHA_3_256, 0, 0},
  {SHA3_384, KIMD, SHA_3_384, 0, 0},
@@ -489,10 +489,10 @@ libica_func_list_element_int icaList[] = {
  {AES_GCM_KMA,  MSA8, AES_128_GCM_ENCRYPT,       0, 0},
  {AES_XTS,      MSA4, AES_128_XTS_ENCRYPT,       0, 0},
  {P_RNG,        ADAPTER, 0, ICA_FLAG_SHW | ICA_FLAG_SW, 0}, // SHW (CPACF) + SW
- {EC_DH,        ADAPTER, 0, 0, 0},
- {EC_DSA_SIGN,	ADAPTER, 0, 0, 0},
- {EC_DSA_VERIFY, ADAPTER, 0, 0, 0},
- {EC_KGEN,      ADAPTER, 0, 0, 0},
+ {EC_DH,        ADAPTER, 0, ICA_FLAG_SW, 0},
+ {EC_DSA_SIGN,	ADAPTER, 0, ICA_FLAG_SW, 0},
+ {EC_DSA_VERIFY, ADAPTER, 0, ICA_FLAG_SW, 0},
+ {EC_KGEN,      ADAPTER, 0, ICA_FLAG_SW, 0},
  {ED25519_KEYGEN, MSA9, SCALAR_MULTIPLY_ED25519, 0, 0},
  {ED25519_SIGN,   MSA9, EDDSA_SIGN_ED25519, 0, 0},
  {ED25519_VERIFY, MSA9, EDDSA_VERIFY_ED25519, 0, 0},
@@ -503,8 +503,8 @@ libica_func_list_element_int icaList[] = {
  {X25519_DERIVE,   MSA9, SCALAR_MULTIPLY_X25519, 0, 0},
  {X448_KEYGEN,   MSA9, SCALAR_MULTIPLY_X448, 0, 0},
  {X448_DERIVE,   MSA9, SCALAR_MULTIPLY_X448, 0, 0},
- {RSA_ME,       ADAPTER, 0, 0, 0},
- {RSA_CRT,      ADAPTER, 0, 0, 0},
+ {RSA_ME,       ADAPTER, 0, ICA_FLAG_SW, 0},
+ {RSA_CRT,      ADAPTER, 0, ICA_FLAG_SW, 0},
  {RSA_KEY_GEN_ME, ADAPTER, 0, ICA_FLAG_SW, 0},  // SW (openssl)
  {RSA_KEY_GEN_CRT, ADAPTER, 0, ICA_FLAG_SW, 0}, // SW (openssl)
 
@@ -598,6 +598,15 @@ int s390_initialize_functionlist()
 				e->property |= ICA_PROPERTY_RSA_ALL;
 			}
 			break;
+		case RSA_KEY_GEN_ME: /* fall-through */
+		case RSA_KEY_GEN_CRT:
+#if defined(ICA_FIPS) && OPENSSL_VERSION_PREREQ(3, 0)
+			if (fips & ICA_FIPS_MODE)
+				e->property |= ICA_PROPERTY_RSA_FIPS;
+			else
+				e->property |= ICA_PROPERTY_RSA_ALL;
+#endif
+			break;
 		default:
 			/* Do nothing. */
 			break;
@@ -667,6 +676,19 @@ int s390_get_functionlist(libica_func_list_element *pmech_list,
 		pmech_list[x].property = 0;
 	}
 #endif /* ICA_FIPS */
+
+#ifdef NO_CPACF
+	pmech_list[x].flags &= ~ICA_FLAG_SHW;
+	/* NO_CPACF also removes sw fallbacks for CPACF based functions, but not
+	 * for card-based functions. */
+	if (pmech_list[x].flags & ICA_FLAG_SW && !(pmech_list[x].flags & ICA_FLAG_DHW)) {
+		pmech_list[x].flags &= ~ICA_FLAG_SW;
+	}
+#endif /* NO_CPACF */
+
+#ifdef NO_SW_FALLBACKS
+	pmech_list[x].flags &= ~ICA_FLAG_SW;
+#endif
   }
   return 0;
 }
