@@ -19,6 +19,7 @@
 #include <openssl/ec.h>
 #include <openssl/ecdh.h>
 #include <openssl/ecdsa.h>
+#include <openssl/evp.h>
 #include <openssl/opensslconf.h>
 #ifdef OPENSSL_FIPS
 #include <openssl/fips.h>
@@ -32,21 +33,15 @@
 #include "icastats.h"
 #include "s390_sha.h"
 
+#if OPENSSL_VERSION_PREREQ(3, 0)
+#include <openssl/core_names.h>
+#include <openssl/param_build.h>
+extern OSSL_LIB_CTX *openssl_libctx;
+#endif
+
 #define CPRBXSIZE (sizeof(struct CPRBX))
 #define PARMBSIZE (2048)
 
-#if OPENSSL_VERSION_NUMBER < 0x10100000L
-#define ECDSA_SIG_get0(sig,pr,ps) \
-do { \
-*(pr)=(sig)->r; \
-*(ps)=(sig)->s; \
-} while (0)
-#define ECDSA_SIG_set0(sig,pr,ps) \
-do { \
-(sig)->r=pr; \
-(sig)->s=ps; \
-} while (0)
-#endif
 
 static int eckeygen_cpacf(ICA_EC_KEY *key);
 static int ecdsa_sign_cpacf(const ICA_EC_KEY *priv, const unsigned char *hash,
@@ -63,6 +58,18 @@ int scalar_mulx_cpacf(unsigned char *res_u,
 		      const unsigned char *u,
 		      int curve_nid);
 
+/**
+ * Since kernel 4.10 the zcrypt device driver has multi domain support and
+ * accepts CPRBs via the ioctl ZSECSENDCPRB with domain addressing 0xFFFF
+ * (AUTOSELECT_DOM in zcrypyt.h). This allows for load balancing between
+ * multiple available crypto cards.
+ */
+typedef enum {
+	dom_addressing_autoselect = 0,
+	dom_addressing_default_domain,
+} dom_addressing_t;
+int dom_addressing = dom_addressing_autoselect;
+
 /**
  * Check if openssl does support this ec curve
  */
@@ -74,100 +81,257 @@ static int is_supported_openssl_curve(int nid)
 	return ptr ? 1 : 0;
 }
 
+#if OPENSSL_VERSION_PREREQ(3, 0)
+static int build_pkey_from_params(OSSL_PARAM_BLD *tmpl, int selection,
+								EVP_PKEY **pkey)
+{
+	OSSL_PARAM *params = NULL;
+	EVP_PKEY_CTX *pctx = NULL;
+	int rc = EINVAL;
+
+	params = OSSL_PARAM_BLD_to_param(tmpl);
+	if (params == NULL) {
+		goto out;
+	}
+
+	pctx = EVP_PKEY_CTX_new_id(EVP_PKEY_EC, NULL);
+	if (pctx == NULL) {
+		goto out;
+	}
+
+	if (!EVP_PKEY_fromdata_init(pctx) ||
+			!EVP_PKEY_fromdata(pctx, pkey, selection, params)) {
+		goto out;
+	}
+
+	EVP_PKEY_CTX_free(pctx);
+	pctx = EVP_PKEY_CTX_new(*pkey, NULL);
+	if (pctx == NULL) {
+		goto out;
+	}
+
+	if (selection & OSSL_KEYMGMT_SELECT_PRIVATE_KEY) {
+		if (EVP_PKEY_check(pctx) != 1) {
+			goto out;
+		}
+	} else {
+		if (EVP_PKEY_public_check(pctx) != 1) {
+			goto out;
+		}
+	}
+
+	rc = 0;
+
+out:
+	if (pctx != NULL)
+		EVP_PKEY_CTX_free(pctx);
+	if (params != NULL)
+		OSSL_PARAM_free(params);
+	if (rc != 0 && *pkey != NULL) {
+		EVP_PKEY_free(*pkey);
+		*pkey = NULL;
+	}
+
+	return rc;
+}
+#endif
+
 /**
- * makes a private EC_KEY.
+ * makes a private EC key from given private value using openssl 3.0.
  */
-static EC_KEY *make_eckey(int nid, const unsigned char *p, size_t plen)
+static EVP_PKEY *make_eckey(int nid, const unsigned char *p, size_t plen)
 {
 	int ok = 0;
-	EC_KEY *k = NULL;
-	BIGNUM *priv = NULL;
-	EC_POINT *pub = NULL;
-	const EC_GROUP *grp;
+	EVP_PKEY *ec_pkey = NULL;
+	EC_GROUP *group = NULL;
+	EC_POINT *point = NULL;
+	BIGNUM *bn_priv = NULL;
+	unsigned char *pub_key = NULL;
+	unsigned int pub_key_len;
+	point_conversion_form_t form;
+#if !OPENSSL_VERSION_PREREQ(3, 0)
+	EC_KEY *ec_key;
+#else
+	OSSL_PARAM_BLD *tmpl = NULL;
+	int rc;
+#endif
 
-	if ((k = EC_KEY_new_by_curve_name(nid)) == NULL) {
+	group = EC_GROUP_new_by_curve_name(nid);
+	if (group == NULL) {
 		goto err;
 	}
 
-	if ((priv = BN_bin2bn(p, plen, NULL)) == NULL) {
+	point = EC_POINT_new(group);
+	if (point == NULL) {
 		goto err;
 	}
 
-	if (!EC_KEY_set_private_key(k, priv)) {
+	bn_priv = BN_bin2bn(p, plen, NULL);
+	if (bn_priv == NULL) {
 		goto err;
 	}
 
-	grp = EC_KEY_get0_group(k);
-	if ((pub = EC_POINT_new(grp)) == NULL) {
+	if (!EC_POINT_mul(group, point, bn_priv, NULL, NULL, NULL)) {
 		goto err;
 	}
 
-	if (!EC_POINT_mul(grp, pub, priv, NULL, NULL, NULL)) {
+	form = EC_GROUP_get_point_conversion_form(group);
+	pub_key_len = EC_POINT_point2buf(group, point, form, &pub_key, NULL);
+	if (pub_key_len == 0) {
 		goto err;
 	}
 
-	if (!EC_KEY_set_public_key(k, pub)) {
+#if !OPENSSL_VERSION_PREREQ(3, 0)
+	ec_key = EC_KEY_new_by_curve_name(nid);
+	if (ec_key == NULL) {
 		goto err;
 	}
+
+	point = EC_POINT_new(EC_KEY_get0_group(ec_key));
+	if (point == NULL) {
+		goto err;
+	}
+
+	if (!EC_KEY_oct2priv(ec_key, p, plen) ||
+		!EC_POINT_mul(EC_KEY_get0_group(ec_key), point, EC_KEY_get0_private_key(ec_key), NULL, NULL, NULL) ||
+		!EC_KEY_set_public_key(ec_key, point) ||
+		!EC_KEY_check_key(ec_key)) {
+		goto err;
+	}
+
+	ec_pkey = EVP_PKEY_new();
+	if (ec_pkey == NULL) {
+		goto err;
+	}
+
+	if (!EVP_PKEY_set1_EC_KEY(ec_pkey, ec_key)) {
+		goto err;
+	}
+
+#else
+
+	tmpl = OSSL_PARAM_BLD_new();
+	if (tmpl == NULL) {
+		goto err;
+	}
+
+	if (!OSSL_PARAM_BLD_push_utf8_string(tmpl, OSSL_PKEY_PARAM_GROUP_NAME, OBJ_nid2sn(nid), 0) ||
+		!OSSL_PARAM_BLD_push_octet_string(tmpl, OSSL_PKEY_PARAM_PUB_KEY, pub_key, pub_key_len) ||
+		!OSSL_PARAM_BLD_push_BN(tmpl, OSSL_PKEY_PARAM_PRIV_KEY, bn_priv)) {
+		goto err;
+	}
+
+	rc = build_pkey_from_params(tmpl, EVP_PKEY_KEYPAIR, &ec_pkey);
+	if (rc != 0) {
+		goto err;
+	}
+#endif
+
 	ok = 1;
 
 err:
-	if (priv)
-		BN_clear_free(priv);
-	if (pub)
-		EC_POINT_free(pub);
+	EC_POINT_free(point);
+	EC_GROUP_free(group);
+	BN_free(bn_priv);
+
+#if !OPENSSL_VERSION_PREREQ(3, 0)
+	// because we use EVP_PKEY_set1_EC_KEY above, free the ec_key here.
+	// If we would use EVP_PKEY_assign_EC_KEY, we would just set it to NULL,
+	// but not free it (reference count).
+	if (ec_key != NULL)
+		EC_KEY_free(ec_key);
+#else
+	if (tmpl)
+		OSSL_PARAM_BLD_free(tmpl);
+#endif
 
 	if (ok)
-		return k;
-	else if (k)
-		EC_KEY_free(k);
+		return ec_pkey;
+	else if (ec_pkey)
+		EVP_PKEY_free(ec_pkey);
 
 	return NULL;
 }
 
 /**
- * makes a public EC_KEY.
+ * makes a public EC key using openssl 3.0.
  */
-static EC_KEY *make_public_eckey(int nid, BIGNUM *x, BIGNUM *y, size_t plen)
+static EVP_PKEY *make_public_eckey(int nid, unsigned char *pubkey, size_t publen)
 {
 	int ok = 0;
-	EC_KEY *k = NULL;
-	EC_POINT *pub = NULL;
-	const EC_GROUP *grp;
+#if !OPENSSL_VERSION_PREREQ(3, 0)
+	EC_KEY *ec_key;
+#else
+	OSSL_PARAM_BLD *tmpl = NULL;
+	int rc;
+#endif
+	EVP_PKEY *ec_pkey = NULL;
+	unsigned char *encoded_pubkey;
+	unsigned int encoded_len = publen + 1;
 
-	(void)plen;	/* suppress unused param warning. XXX remove plen? */
-
-	k = EC_KEY_new_by_curve_name(nid);
-	if (!k)
+	encoded_pubkey = OPENSSL_zalloc(encoded_len);
+	if (encoded_pubkey == NULL) {
 		goto err;
+	}
+	encoded_pubkey[0] = 0x04;
+	memcpy(encoded_pubkey + 1, pubkey, publen);
 
-	grp = EC_KEY_get0_group(k);
-	pub = EC_POINT_new(grp);
-	if (!pub)
+#if !OPENSSL_VERSION_PREREQ(3, 0)
+
+	ec_key = EC_KEY_new_by_curve_name(nid);
+	if (ec_key == NULL) {
 		goto err;
-
-	if (x && y) {
-		BN_CTX* ctx = BN_CTX_new();
-		if (ctx == NULL)
-			goto err;
-
-		EC_POINT_set_affine_coordinates_GFp(grp, pub, x, y, ctx);
-
-		BN_CTX_free(ctx);
 	}
 
-	if (!EC_KEY_set_public_key(k, pub))
+	if (!EC_KEY_oct2key(ec_key, encoded_pubkey, encoded_len, NULL) ||
+		!EC_KEY_check_key(ec_key)) {
 		goto err;
+	}
+
+	ec_pkey = EVP_PKEY_new();
+	if (ec_pkey == NULL) {
+		goto err;
+	}
+
+	if (!EVP_PKEY_set1_EC_KEY(ec_pkey, ec_key)) {
+		goto err;
+	}
+
+#else
+
+	tmpl = OSSL_PARAM_BLD_new();
+	if (tmpl == NULL) {
+		goto err;
+	}
+
+	if (!OSSL_PARAM_BLD_push_utf8_string(tmpl, OSSL_PKEY_PARAM_GROUP_NAME, OBJ_nid2sn(nid), 0) ||
+		!OSSL_PARAM_BLD_push_octet_string(tmpl, OSSL_PKEY_PARAM_PUB_KEY, encoded_pubkey, encoded_len)) {
+		goto err;
+	}
+
+	rc = build_pkey_from_params(tmpl, EVP_PKEY_PUBLIC_KEY, &ec_pkey);
+	if (rc != 0) {
+		goto err;
+	}
+#endif
+
 	ok = 1;
 
 err:
-	if (pub)
-		EC_POINT_free(pub);
+	if (encoded_pubkey != NULL)
+		OPENSSL_free(encoded_pubkey);
+#if !OPENSSL_VERSION_PREREQ(3, 0)
+	if (ec_key)
+		EC_KEY_free(ec_key);
+#else
+	if (tmpl)
+		OSSL_PARAM_BLD_free(tmpl);
+#endif
 
 	if (ok)
-		return k;
-	else if (k)
-		EC_KEY_free(k);
+		return ec_pkey;
+	else if (ec_pkey)
+		EVP_PKEY_free(ec_pkey);
 
 	return NULL;
 }
@@ -248,8 +412,10 @@ static unsigned int make_cprbx(struct CPRBX* cprbx, unsigned int parmlen,
 	cprbx->cprb_ver_id = 0x02;
 	memcpy(&(cprbx->func_id), "T2", 2);
 	cprbx->req_parml = parmlen;
-	cprbx->domain = get_default_domain();
-
+	if (dom_addressing == dom_addressing_autoselect)
+		cprbx->domain = 0xFFFF;
+	else
+		cprbx->domain = get_default_domain();
 	cprbx->rpl_msgbl = CPRBXSIZE + PARMBSIZE;
 	cprbx->req_parmb = ((uint8_t *) preqcblk) + CPRBXSIZE;
 	cprbx->rpl_parmb = ((uint8_t *) prepcblk) + CPRBXSIZE;
@@ -739,7 +905,7 @@ unsigned int ecdh_hw(ica_adapter_handle_t adapter_handle,
 		const ICA_EC_KEY *privkey_A, const ICA_EC_KEY *pubkey_B,
 		unsigned char *z)
 {
-	uint8_t *buf;
+	uint8_t *buf = NULL;
 	size_t len;
 	int rc;
 	struct ica_xcRB xcrb;
@@ -753,20 +919,33 @@ unsigned int ecdh_hw(ica_adapter_handle_t adapter_handle,
 			return rc;
 	}
 
-	if (!ecc_via_online_card)
+	if (privkey_A->nid != pubkey_B->nid ||
+		!curve_supported_via_online_card(privkey_A->nid))
 		return ENODEV;
 
 	if (adapter_handle == DRIVER_NOT_LOADED)
 		return EIO;
 
 	reply_p = make_ecdh_request(privkey_A, pubkey_B, &xcrb, &buf, &len);
-	if (!reply_p)
-		return EIO;
+	if (!reply_p) {
+		rc = EIO;
+		goto ret;
+	}
 
 	rc = ioctl(adapter_handle, ZSECSENDCPRB, xcrb);
 	if (rc != 0) {
-		rc = EIO;
-		goto ret;
+		dom_addressing = dom_addressing_default_domain;
+		reply_p = make_ecdh_request(privkey_A, pubkey_B, &xcrb, &buf, &len);
+		if (!reply_p) {
+			rc = EIO;
+			goto ret;
+		}
+
+		rc = ioctl(adapter_handle, ZSECSENDCPRB, xcrb);
+		if (rc != 0) {
+			rc = EIO;
+			goto ret;
+		}
 	}
 
 	if (reply_p->key_len - 4 != privlen) {
@@ -777,8 +956,10 @@ unsigned int ecdh_hw(ica_adapter_handle_t adapter_handle,
 	memcpy(z, reply_p->raw_z_value, privlen);
 	rc = 0;
 ret:
-	OPENSSL_cleanse(buf, len);
-	free(buf);
+	if (buf) {
+		OPENSSL_cleanse(buf, len);
+		free(buf);
+	}
 	return rc;
 }
 
@@ -792,53 +973,58 @@ ret:
 unsigned int ecdh_sw(const ICA_EC_KEY *privkey_A, const ICA_EC_KEY *pubkey_B,
 		unsigned char *z)
 {
-	int rc, ret = EIO;
-	EC_KEY *a = NULL;
-	EC_KEY *b = NULL;
-	BIGNUM *xb, *yb;
-	unsigned int ztmplen;
-	unsigned int privlen = privlen_from_nid(privkey_A->nid);
+	int ret = 0;
+	EVP_PKEY *a = NULL, *b = NULL;
+	EVP_PKEY_CTX *ctx = NULL;
+	size_t privlen = privlen_from_nid(privkey_A->nid);
 
 #ifdef ICA_FIPS
-	if ((fips & ICA_FIPS_MODE) && (!FIPS_mode()))
-		return EACCES;
+	if ((fips & ICA_FIPS_MODE) && (!openssl_in_fips_mode()))
+	return EACCES;
 #endif /* ICA_FIPS */
 
-	if (!is_supported_openssl_curve(privkey_A->nid))
-		return EINVAL;
+	BEGIN_OPENSSL_LIBCTX(openssl_libctx, ret);
+
+	if (!is_supported_openssl_curve(privkey_A->nid)) {
+		ret = EPERM;
+		goto err;
+	}
 
 	a = make_eckey(privkey_A->nid, privkey_A->D, privlen);
-	xb = BN_bin2bn(pubkey_B->X, privlen, NULL);
-	yb = BN_bin2bn(pubkey_B->Y, privlen, NULL);
-	b = make_public_eckey(privkey_A->nid, xb, yb, privlen);
-	if (!a || !b)
+	b = make_public_eckey(pubkey_B->nid, pubkey_B->X, 2 * privlen);
+	if (!a || !b) {
+		ret = EIO;
 		goto err;
+	}
 
-	ztmplen = (EC_GROUP_get_degree(EC_KEY_get0_group(a)) + 7) / 8;
-	if (ztmplen != privlen)
+	ctx = EVP_PKEY_CTX_new(a, NULL);
+	if (ctx == NULL) {
+		ret = EIO;
 		goto err;
+	}
 
-	/**
-	 * Make sure to use original openssl compute_key method to avoid endless loop
-	 * when being called from IBMCA engine in software fallback.
-	 */
-#if OPENSSL_VERSION_NUMBER < 0x10100000L
-	ECDH_set_method(a, ECDH_OpenSSL());
-#else
-	EC_KEY_set_method(a, EC_KEY_OpenSSL());
-#endif
-	rc = ECDH_compute_key(z, privlen, EC_KEY_get0_public_key(b), a, NULL);
-	if (rc == 0)
+	if (EVP_PKEY_derive_init(ctx) <= 0 ||
+			EVP_PKEY_derive_set_peer(ctx, b) <= 0) {
+		ret = EIO;
 		goto err;
+	}
+
+	if (EVP_PKEY_derive(ctx, z, &privlen) <= 0) {
+		ret = EIO;
+		goto err;
+	}
 
 	ret = 0;
 
 err:
-	BN_clear_free(xb);
-	BN_clear_free(yb);
-	EC_KEY_free(a);
-	EC_KEY_free(b);
+	if (a != NULL)
+		EVP_PKEY_free(a);
+	if (b != NULL)
+		EVP_PKEY_free(b);
+	if (ctx != NULL)
+		EVP_PKEY_CTX_free(ctx);
 
+	END_OPENSSL_LIBCTX(ret);
 	return ret;
 }
 
@@ -1042,12 +1228,18 @@ static ECDSA_SIGN_REPLY* make_ecdsa_sign_request(const ICA_EC_KEY *privkey,
  */
 static unsigned int provide_pubkey(const ICA_EC_KEY *privkey, unsigned char *X, unsigned char *Y)
 {
-	EC_KEY *eckey = NULL;
-	EC_POINT *pub_key = NULL;
+#if !OPENSSL_VERSION_PREREQ(3, 0)
 	const EC_GROUP *group = NULL;
+	EC_KEY *eckey = NULL;
+#else
+	EVP_PKEY *eckey = NULL;
+	unsigned char *ecpoint = NULL;
+	size_t ecpoint_len;
+#endif
+	EC_POINT *pub_key = NULL;
 	BIGNUM *bn_d = NULL, *bn_x = NULL, *bn_y = NULL;
 	int privlen = -1;
-	unsigned int i, n, rc;
+	unsigned int rc;
 
 	if (privkey == NULL || X == NULL || Y == NULL) {
 		return EFAULT;
@@ -1071,6 +1263,7 @@ static unsigned int provide_pubkey(const ICA_EC_KEY *privkey, unsigned char *X,
 	}
 
 	/* Calculate public (X,Y) values */
+#if !OPENSSL_VERSION_PREREQ(3, 0)
 	eckey = EC_KEY_new_by_curve_name(privkey->nid);
 	EC_KEY_set_private_key(eckey, bn_d);
 	group = EC_KEY_get0_group(eckey);
@@ -1088,26 +1281,50 @@ static unsigned int provide_pubkey(const ICA_EC_KEY *privkey, unsigned char *X,
 		goto end;
 	}
 
-	/* Format (X) as char array, with leading zeros if necessary */
-	n = privlen - BN_num_bytes(bn_x);
-	for (i = 0; i < n; i++)
-		X[i] = 0x00;
-	BN_bn2bin(bn_x, &(X[n]));
+	/* Format (X,Y) as char array, with leading zeros if necessary */
+	BN_bn2binpad(bn_x, X, privlen);
+	BN_bn2binpad(bn_y, Y, privlen);
 
-	/* Format (Y) as char array, with leading zeros if necessary */
-	n = privlen - BN_num_bytes(bn_y);
-	for (i = 0; i < n; i++)
-		Y[i] = 0x00;
-	BN_bn2bin(bn_y, &(Y[n]));
+#else
+	eckey = make_eckey(privkey->nid, privkey->D, privlen);
+	if (eckey == NULL) {
+		goto end;
+	}
+
+	if (!EVP_PKEY_get_octet_string_param(eckey, OSSL_PKEY_PARAM_ENCODED_PUBLIC_KEY,
+									NULL, 0, &ecpoint_len)) {
+		goto end;
+	}
+
+	ecpoint = OPENSSL_zalloc(ecpoint_len);
+	if (ecpoint == NULL) {
+		goto end;
+	}
+
+	if (!EVP_PKEY_get_octet_string_param(eckey, OSSL_PKEY_PARAM_ENCODED_PUBLIC_KEY,
+									ecpoint, ecpoint_len, &ecpoint_len)) {
+		goto end;
+	}
+
+	memcpy(X, ecpoint + 1, privlen); // skip leading 0x04
+	memcpy(Y, ecpoint + 1 + privlen, privlen);
+
+#endif
 
 	rc = 0;
 
 end:
-
 	if (pub_key)
 		EC_POINT_free(pub_key);
+#if !OPENSSL_VERSION_PREREQ(3, 0)
 	if (eckey)
 		EC_KEY_free(eckey);
+#else
+	if (eckey)
+		EVP_PKEY_free(eckey);
+	if (ecpoint)
+		OPENSSL_free(ecpoint);
+#endif
 	BN_clear_free(bn_x);
 	BN_clear_free(bn_y);
 	BN_clear_free(bn_d);
@@ -1125,7 +1342,7 @@ unsigned int ecdsa_sign_hw(ica_adapter_handle_t adapter_handle,
 		const ICA_EC_KEY *privkey, const unsigned char *hash, unsigned int hash_length,
 		unsigned char *signature)
 {
-	uint8_t *buf;
+	uint8_t *buf = NULL;
 	size_t len;
 	int rc;
 	struct ica_xcRB xcrb;
@@ -1141,7 +1358,7 @@ unsigned int ecdsa_sign_hw(ica_adapter_handle_t adapter_handle,
 			return rc;
 	}
 
-	if (!ecc_via_online_card)
+	if (!curve_supported_via_online_card(privkey->nid))
 		return ENODEV;
 
 	if (adapter_handle == DRIVER_NOT_LOADED)
@@ -1153,13 +1370,26 @@ unsigned int ecdsa_sign_hw(ica_adapter_handle_t adapter_handle,
 
 	reply_p = make_ecdsa_sign_request((const ICA_EC_KEY*)privkey,
 			X, Y, hash, hash_length, &xcrb, &buf, &len);
-	if (!reply_p)
-		return EIO;
+	if (!reply_p) {
+		rc = EIO;
+		goto ret;
+	}
 
 	rc = ioctl(adapter_handle, ZSECSENDCPRB, xcrb);
 	if (rc != 0) {
-		rc = EIO;
-		goto ret;
+		dom_addressing = dom_addressing_default_domain;
+		reply_p = make_ecdsa_sign_request((const ICA_EC_KEY*)privkey,
+				X, Y, hash, hash_length, &xcrb, &buf, &len);
+		if (!reply_p) {
+			rc = EIO;
+			goto ret;
+		}
+
+		rc = ioctl(adapter_handle, ZSECSENDCPRB, xcrb);
+		if (rc != 0) {
+			rc = EIO;
+			goto ret;
+		}
 	}
 
 	if (reply_p->vud_len - 8 != 2 * privlen) {
@@ -1170,8 +1400,10 @@ unsigned int ecdsa_sign_hw(ica_adapter_handle_t adapter_handle,
 	memcpy(signature, reply_p->signature, reply_p->vud_len-8);
 	rc = 0;
 ret:
-	OPENSSL_cleanse(buf, len);
-	free(buf);
+	if (buf) {
+		OPENSSL_cleanse(buf, len);
+		free(buf);
+	}
 	return rc;
 }
 
@@ -1184,61 +1416,88 @@ unsigned int ecdsa_sign_sw(const ICA_EC_KEY *privkey,
 		const unsigned char *hash, unsigned int hash_length,
 		unsigned char *signature)
 {
-	int rc = EIO;
-	EC_KEY *a = NULL;
-	BIGNUM* r=NULL; BIGNUM* s=NULL;
-	ECDSA_SIG* sig = NULL;
+	int rc = 0;
+    EVP_PKEY *ec_pkey = NULL;
+    ECDSA_SIG *sig = NULL;
+    const BIGNUM *r, *s;
+    EVP_PKEY_CTX *ctx = NULL;
+    size_t siglen;
+    unsigned char *sigbuf = NULL;
+    const unsigned char *p;
 	unsigned int privlen = privlen_from_nid(privkey->nid);
-	unsigned int i,n;
 
 #ifdef ICA_FIPS
-	if ((fips & ICA_FIPS_MODE) && (!FIPS_mode()))
+	if ((fips & ICA_FIPS_MODE) && (!openssl_in_fips_mode()))
 		return EACCES;
 #endif /* ICA_FIPS */
 
-	if (!is_supported_openssl_curve(privkey->nid))
-		return EINVAL;
+	BEGIN_OPENSSL_LIBCTX(openssl_libctx, rc);
 
-	a = make_eckey(privkey->nid, privkey->D, privlen);
-	if (!a)
+	if (!is_supported_openssl_curve(privkey->nid)) {
+		rc = EPERM;
 		goto err;
+	}
 
-	if (!EC_KEY_check_key(a))
+	ec_pkey = make_eckey(privkey->nid, privkey->D, privlen);
+	if (ec_pkey == NULL) {
+		rc = EIO;
 		goto err;
+	}
 
-	/**
-	 * Make sure to use original openssl sign method to avoid endless loop when being
-	 * called from IBMCA engine in software fallback.
-	 */
-#if OPENSSL_VERSION_NUMBER < 0x10100000L
-	ECDSA_set_method(a, ECDSA_OpenSSL());
-#else
-	EC_KEY_set_method(a, EC_KEY_OpenSSL());
-#endif
-	sig = ECDSA_do_sign(hash, hash_length, a);
-	if (!sig)
+	ctx = EVP_PKEY_CTX_new(ec_pkey, NULL);
+	if (ctx == NULL) {
+		rc = EIO;
 		goto err;
+	}
 
-	ECDSA_SIG_get0(sig, (const BIGNUM**)&r, (const BIGNUM **)&s);
+	if (EVP_PKEY_sign_init(ctx) <= 0) {
+		rc = EIO;
+		goto err;
+	}
+
+	if (EVP_PKEY_sign(ctx, NULL, &siglen, hash, (size_t)hash_length) <= 0) {
+		rc = EIO;
+		goto err;
+	}
+
+	sigbuf = malloc(siglen);
+	if (sigbuf == NULL) {
+		rc = EIO;
+		goto err;
+	}
+
+	if (EVP_PKEY_sign(ctx, sigbuf, &siglen, hash, (size_t)hash_length) <= 0) {
+		rc = EIO;
+		goto err;
+	}
+
+	p = sigbuf;
+	sig = d2i_ECDSA_SIG(NULL, &p, siglen);
+	if (sig == NULL) {
+		rc = EIO;
+		goto err;
+	}
+
+	ECDSA_SIG_get0(sig, &r, &s);
 
 	/* Insert leading 0x00's if r or s shorter than privlen */
-	n = privlen - BN_num_bytes(r);
-	for (i = 0; i < n; i++)
-		signature[i] = 0x00;
-	BN_bn2bin(r, &(signature[n]));
-
-	n = privlen - BN_num_bytes(s);
-	for (i = 0; i < n; i++)
-		signature[privlen+i] = 0x00;
-	BN_bn2bin(s, &(signature[privlen+n]));
+	BN_bn2binpad(r, signature, privlen);
+	BN_bn2binpad(s, signature + privlen, privlen);
 
 	rc = 0;
 
 err:
-	ECDSA_SIG_free(sig); /* also frees r and s */
-	EC_KEY_free(a);
+	if (sig != NULL)
+		ECDSA_SIG_free(sig);
+	if (ec_pkey != NULL)
+		EVP_PKEY_free(ec_pkey);
+	if (sigbuf != NULL)
+		free(sigbuf);
+	if (ctx != NULL)
+		EVP_PKEY_CTX_free(ctx);
 
-	return (rc);
+	END_OPENSSL_LIBCTX(rc);
+	return rc;
 }
 
 /**
@@ -1582,7 +1841,7 @@ unsigned int ecdsa_verify_hw(ica_adapter_handle_t adapter_handle,
 		const ICA_EC_KEY *pubkey, const unsigned char *hash, unsigned int hash_length,
 		const unsigned char *signature)
 {
-	uint8_t *buf;
+	uint8_t *buf = NULL;
 	size_t len;
 	int rc;
 	struct ica_xcRB xcrb;
@@ -1594,7 +1853,7 @@ unsigned int ecdsa_verify_hw(ica_adapter_handle_t adapter_handle,
 			return rc;
 	}
 
-	if (!ecc_via_online_card)
+	if (!curve_supported_via_online_card(pubkey->nid))
 		return ENODEV;
 
 	if (adapter_handle == DRIVER_NOT_LOADED)
@@ -1602,13 +1861,26 @@ unsigned int ecdsa_verify_hw(ica_adapter_handle_t adapter_handle,
 
 	reply_p = make_ecdsa_verify_request(pubkey, hash, hash_length,
 					    signature, &xcrb, &buf, &len);
-	if (!reply_p)
-		return EIO;
+	if (!reply_p) {
+		rc = EIO;
+		goto ret;
+	}
 
 	rc = ioctl(adapter_handle, ZSECSENDCPRB, xcrb);
 	if (rc != 0) {
-		rc = EIO;
-		goto ret;
+		dom_addressing = dom_addressing_default_domain;
+		reply_p = make_ecdsa_verify_request(pubkey, hash, hash_length,
+						    signature, &xcrb, &buf, &len);
+		if (!reply_p) {
+			rc = EIO;
+			goto ret;
+		}
+
+		rc = ioctl(adapter_handle, ZSECSENDCPRB, xcrb);
+		if (rc != 0) {
+			rc = EIO;
+			goto ret;
+		}
 	}
 
 	if (((struct CPRBX*)reply_p)->ccp_rtcode == 4 &&
@@ -1625,8 +1897,10 @@ unsigned int ecdsa_verify_hw(ica_adapter_handle_t adapter_handle,
 
 	rc = 0;
 ret:
-	OPENSSL_cleanse(buf, len);
-	free(buf);
+	if (buf) {
+		OPENSSL_cleanse(buf, len);
+		free(buf);
+	}
 	return rc;
 }
 
@@ -1640,44 +1914,69 @@ ret:
 unsigned int ecdsa_verify_sw(const ICA_EC_KEY *pubkey,
 		const unsigned char *hash, unsigned int hash_length,
 		const unsigned char *signature) {
-	int rc = EIO;
-	EC_KEY *a = NULL;
-	BIGNUM *r, *s, *xa, *ya;
-	ECDSA_SIG* sig = NULL;
+	int rc = 0;
+	BIGNUM *r = NULL, *s = NULL;
+	ECDSA_SIG *sig = NULL;
+	unsigned char *sigbuf = NULL;
+	EVP_PKEY_CTX *ctx = NULL;
+	size_t siglen;
+	EVP_PKEY *ec_pkey = NULL;
 	unsigned int privlen = privlen_from_nid(pubkey->nid);
 
 #ifdef ICA_FIPS
-	if ((fips & ICA_FIPS_MODE) && (!FIPS_mode()))
+	if ((fips & ICA_FIPS_MODE) && (!openssl_in_fips_mode()))
 	return EACCES;
 #endif /* ICA_FIPS */
 
-	if (!is_supported_openssl_curve(pubkey->nid))
-		return EINVAL;
+	BEGIN_OPENSSL_LIBCTX(openssl_libctx, rc);
 
-	/* create public key with given (x,y) */
-	xa = BN_bin2bn(pubkey->X, privlen, NULL);
-	ya = BN_bin2bn(pubkey->Y, privlen, NULL);
-	a = make_public_eckey(pubkey->nid, xa, ya, privlen);
-	if (!a) {
+	if (!is_supported_openssl_curve(pubkey->nid)) {
+		rc = EINVAL;
 		goto err;
 	}
 
-	/* create ECDSA_SIG instance */
 	sig = ECDSA_SIG_new();
+	if (sig == NULL) {
+		rc = EIO;
+		goto err;
+	}
+
 	r = BN_bin2bn(signature, privlen, NULL);
 	s = BN_bin2bn(signature + privlen, privlen, NULL);
-	ECDSA_SIG_set0(sig, r, s);
+	if (r == NULL || s == NULL) {
+		rc = EIO;
+		goto err;
+	}
 
-	/**
-	 * Make sure to use original openssl verify method to avoid endless loop
-	 * when being called from IBMCA engine in software fallback.
-	 */
-#if OPENSSL_VERSION_NUMBER < 0x10100000L
-	ECDSA_set_method(a, ECDSA_OpenSSL());
-#else
-	EC_KEY_set_method(a, EC_KEY_OpenSSL());
-#endif
-	rc = ECDSA_do_verify(hash, hash_length, sig, a);
+	if (!ECDSA_SIG_set0(sig, r, s)) {
+		rc = EIO;
+		goto err;
+	}
+
+	siglen = i2d_ECDSA_SIG(sig, &sigbuf);
+	if (siglen <= 0) {
+		rc = EIO;
+		goto err;
+	}
+
+	ec_pkey = make_public_eckey(pubkey->nid, pubkey->X, 2 * privlen);
+	if (ec_pkey == NULL) {
+		rc = EIO;
+		goto err;
+	}
+
+	ctx = EVP_PKEY_CTX_new(ec_pkey, NULL);
+	if (ctx == NULL) {
+		rc = EIO;
+		goto err;
+	}
+
+	if (EVP_PKEY_verify_init(ctx) <= 0) {
+		rc = EIO;
+		goto err;
+	}
+
+	rc = EVP_PKEY_verify(ctx, sigbuf, siglen, hash, hash_length);
 	switch (rc) {
 	case 0: /* signature invalid */
 		rc = EFAULT;
@@ -1691,11 +1990,16 @@ unsigned int ecdsa_verify_sw(const ICA_EC_KEY *pubkey,
 	}
 
 err:
-	BN_clear_free(xa);
-	BN_clear_free(ya);
-	ECDSA_SIG_free(sig);
-	EC_KEY_free(a);
+	if (sig != NULL)
+		ECDSA_SIG_free(sig);
+	if (ec_pkey != NULL)
+		EVP_PKEY_free(ec_pkey);
+	if (sigbuf != NULL)
+		OPENSSL_free(sigbuf);
+	if (ctx != NULL)
+		EVP_PKEY_CTX_free(ctx);
 
+	END_OPENSSL_LIBCTX(rc);
 	return rc;
 }
 
@@ -1940,7 +2244,7 @@ static int eckeygen_cpacf(ICA_EC_KEY *key)
  */
 unsigned int eckeygen_hw(ica_adapter_handle_t adapter_handle, ICA_EC_KEY *key)
 {
-	uint8_t *buf;
+	uint8_t *buf = NULL;
 	size_t len;
 	int rc;
 	struct ica_xcRB xcrb;
@@ -1955,17 +2259,29 @@ unsigned int eckeygen_hw(ica_adapter_handle_t adapter_handle, ICA_EC_KEY *key)
 			return rc;
 	}
 
-	if (!ecc_via_online_card)
+	if (!curve_supported_via_online_card(key->nid))
 		return ENODEV;
 
 	reply_p = make_eckeygen_request(key, &xcrb, &buf, &len);
-	if (!reply_p)
-		return EIO;
+	if (!reply_p) {
+		rc = EIO;
+		goto ret;
+	}
 
 	rc = ioctl(adapter_handle, ZSECSENDCPRB, xcrb);
 	if (rc != 0) {
-		rc = EIO;
-		goto ret;
+		dom_addressing = dom_addressing_default_domain;
+		reply_p = make_eckeygen_request(key, &xcrb, &buf, &len);
+		if (!reply_p) {
+			rc = EIO;
+			goto ret;
+		}
+
+		rc = ioctl(adapter_handle, ZSECSENDCPRB, xcrb);
+		if (rc != 0) {
+			rc = EIO;
+			goto ret;
+		}
 	}
 
 	if (reply_p->eckey.privsec.formatted_data_len != privlen) {
@@ -1985,8 +2301,10 @@ unsigned int eckeygen_hw(ica_adapter_handle_t adapter_handle, ICA_EC_KEY *key)
 	memcpy(key->X, (char*)pub_p->pubkey, 2*privlen);
 	rc = 0;
 ret:
-	OPENSSL_cleanse(buf, len);
-	free(buf);
+	if (buf) {
+		OPENSSL_cleanse(buf, len);
+		free(buf);
+	}
 	return rc;
 }
 
@@ -1998,81 +2316,179 @@ ret:
  */
 unsigned int eckeygen_sw(ICA_EC_KEY *key)
 {
-	int rc = EIO;
-	EC_KEY *a = NULL;
-	BIGNUM* d=NULL; BIGNUM *x=NULL; BIGNUM *y=NULL;
-	const EC_POINT* q=NULL;
-	const EC_GROUP *group=NULL;
+#if !OPENSSL_VERSION_PREREQ(3, 0)
+	const EC_KEY *ec_key = NULL;
+	BN_CTX *bnctx = NULL;
+	size_t d_len;
+#else
+	BIGNUM *bn_d = NULL;
+#endif
+	int rc = 0;
+	EVP_PKEY_CTX *ctx = NULL;
+	EVP_PKEY *ec_pkey = NULL;
+	unsigned char *ecpoint = NULL, *d = NULL;
+	size_t ecpoint_len;
 	unsigned int privlen = privlen_from_nid(key->nid);
-	unsigned int i, n;
 
 #ifdef ICA_FIPS
-	if ((fips & ICA_FIPS_MODE) && (!FIPS_mode()))
+	if ((fips & ICA_FIPS_MODE) && (!openssl_in_fips_mode()))
 		return EACCES;
 #endif /* ICA_FIPS */
 
-	if (!is_supported_openssl_curve(key->nid))
-		return EINVAL;
+	BEGIN_OPENSSL_LIBCTX(openssl_libctx, rc);
 
-	if ((a = EC_KEY_new_by_curve_name(key->nid)) == NULL)
+	if (!is_supported_openssl_curve(key->nid)) {
+		rc = EPERM;
 		goto err;
+	}
 
-	if ((group = EC_KEY_get0_group(a)) == NULL)
+	ctx = EVP_PKEY_CTX_new_id(EVP_PKEY_EC, NULL);
+	if (ctx == NULL) {
+		rc = EIO;
 		goto err;
+	}
 
-#if OPENSSL_VERSION_NUMBER < 0x10100000L
-	/* Openssl 1.0.2 does not provide a default ec_key_gen method, so IBMCA does not
-	 * override such a function. Therefore nothing to do here.
-	 */
+	if (EVP_PKEY_keygen_init(ctx) <= 0 ||
+		EVP_PKEY_CTX_set_ec_paramgen_curve_nid(ctx, key->nid) <= 0 ||
+		EVP_PKEY_keygen(ctx, &ec_pkey) <= 0) {
+		rc = EIO;
+		goto err;
+	}
+
+#if !OPENSSL_VERSION_PREREQ(3, 0)
+	ec_key = EVP_PKEY_get0_EC_KEY(ec_pkey);
+	if (ec_key == NULL) {
+		rc = EIO;
+		goto err;
+	}
+
+	bnctx = BN_CTX_new();
+	if (bnctx == NULL) {
+		rc = ENOMEM;
+		goto err;
+	}
+
+	ecpoint_len = EC_KEY_key2buf(ec_key, POINT_CONVERSION_UNCOMPRESSED,
+								&ecpoint, bnctx);
+	if (ecpoint_len == 0) {
+		rc = EIO;
+		goto err;
+	}
+
+	memcpy(key->X, ecpoint + 1, privlen); // skip leading 0x04
+	memcpy(key->Y, ecpoint + 1 + privlen, privlen);
+
+	d_len = EC_KEY_priv2buf(ec_key, &d);
+	if (d_len != privlen) {
+		rc = EIO;
+		goto err;
+	}
+	memcpy(key->D, d, d_len);
 #else
-	/**
-	 * IBMCA overrides the default ec_key_gen method for openssl 1.1.0 and later.
-	 * Make sure to use original openssl method to avoid endless loop when being
-	 * called from IBMCA engine in software fallback.
-	 */
-	EC_KEY_set_method(a, EC_KEY_OpenSSL());
+
+	/* Provide public key (X,Y) */
+	if (!EVP_PKEY_get_octet_string_param(ec_pkey,
+					OSSL_PKEY_PARAM_ENCODED_PUBLIC_KEY,
+					NULL, 0, &ecpoint_len)) {
+		rc = EIO;
+		goto err;
+	}
+
+	ecpoint = OPENSSL_zalloc(ecpoint_len);
+	if (ecpoint == NULL) {
+		rc = EIO;
+		goto err;
+	}
+
+	if (!EVP_PKEY_get_octet_string_param(ec_pkey,
+					OSSL_PKEY_PARAM_ENCODED_PUBLIC_KEY,
+					ecpoint, ecpoint_len, &ecpoint_len)) {
+		rc = EIO;
+		goto err;
+	}
+
+	memcpy(key->X, ecpoint + 1, privlen); // skip leading 0x04
+	memcpy(key->Y, ecpoint + 1 + privlen, privlen);
+
+	/* Provide private key (D) */
+	if (!EVP_PKEY_get_bn_param(ec_pkey, OSSL_PKEY_PARAM_PRIV_KEY, &bn_d)) {
+		rc = EIO;
+		goto err;
+	}
+	BN_bn2binpad(bn_d, key->D, privlen);
 #endif
 
-	if (!EC_KEY_generate_key(a))
-		goto err;
-
-	/* provide private key */
-	d = (BIGNUM*)EC_KEY_get0_private_key(a);
-	n = privlen - BN_num_bytes(d);
-	for (i=0;i<n;i++)
-		key->D[i] = 0x00;
-	BN_bn2bin(d, &(key->D[n]));
-
-	/* provide public key */
-	q = EC_KEY_get0_public_key(a);
-	x = BN_new();
-	y = BN_new();
-	if (!EC_POINT_get_affine_coordinates_GFp(group, q, x, y, NULL))
-		goto err;
-
-	/* pub(X) */
-	n = privlen - BN_num_bytes(x);
-	for (i=0; i<n; i++)
-		key->X[i] = 0x00;
-	BN_bn2bin(x, &(key->X[n]));
-
-	/* pub(Y) */
-	n = privlen - BN_num_bytes(y);
-	for (i=0; i<n; i++)
-		key->Y[i] = 0x00;
-	BN_bn2bin(y, &(key->Y[n]));
-
 	rc = 0;
 
 err:
-	/* cleanup */
-	EC_KEY_free(a); /* also frees d */
-	BN_clear_free(x);
-	BN_clear_free(y);
+	if (ctx != NULL)
+		EVP_PKEY_CTX_free(ctx);
+#if !OPENSSL_VERSION_PREREQ(3, 0)
+	if (bnctx != NULL)
+		BN_CTX_free(bnctx);
+#else
+	if (bn_d != NULL)
+		BN_free(bn_d);
+#endif
+	if (ec_pkey != NULL)
+		EVP_PKEY_free(ec_pkey);
+	if (ecpoint != NULL)
+		OPENSSL_free(ecpoint);
+	if (d != NULL)
+		OPENSSL_free(d);
 
+	END_OPENSSL_LIBCTX(rc);
 	return rc;
 }
 
+/**
+ * Check if the given ICA_EC_KEY is valid. This check is performed via openssl,
+ * so we can only check keys that are supported by openssl. If e.g. openssl
+ * is in fips mode, very few curves are supported.
+ *
+ * @return:
+ *    0       success
+ *    EINVAL  key check failed
+ */
+int ec_key_check(ICA_EC_KEY *ica_key)
+{
+	EVP_PKEY *privkey = NULL, *pubkey = NULL;
+	BIGNUM *d = NULL, *x = NULL, *y = NULL;
+	int privlen, rc = EINVAL;
+
+	if (!ica_key)
+		goto done;
+
+	privlen = privlen_from_nid(ica_key->nid);
+
+	d = BN_bin2bn(ica_key->D, privlen, NULL);
+	if (!BN_is_zero(d)) {
+		privkey = make_eckey(ica_key->nid, ica_key->D, privlen);
+		if (!privkey)
+			goto done;
+	}
+
+	x = BN_bin2bn(ica_key->X, privlen, NULL);
+	y = BN_bin2bn(ica_key->Y, privlen, NULL);
+	if (!BN_is_zero(x) && !BN_is_zero(y)) {
+		pubkey = make_public_eckey(ica_key->nid, ica_key->X, 2 * privlen);
+		if (!pubkey)
+			goto done;
+	}
+
+	rc = 0;
+
+done:
+	BN_clear_free(d);
+	BN_clear_free(x);
+	BN_clear_free(y);
+	if (pubkey)
+		EVP_PKEY_free(pubkey);
+	if (privkey)
+		EVP_PKEY_free(privkey);
+
+	return rc;
+}
 
 /*
  * Derive public key.
@@ -2251,6 +2667,7 @@ do {									    \
 	exit(TEST_FAIL);						    \
 } while(0)
 
+#ifndef NO_CPACF
 static void ecdsa_test(void)
 {
 	unsigned long long rnd[2];
@@ -2298,6 +2715,7 @@ static void ecdsa_test(void)
 			TEST_ERROR("Unknown hash", "ECDSA", i);
 		}
 
+
 		if (rc)
 			TEST_ERROR("Hashing failed", "ECDSA", i);
 
@@ -2639,9 +3057,14 @@ static void scalar_mul_test(void)
 		t4++;
 	}
 }
+#endif /* NO_CPACF */
 
 int main(void)
 {
+#ifdef NO_CPACF
+	printf("Skipping EC internal test, because CPACF support disabled via config option.\n");
+	exit(TEST_SKIP);
+#else
 	if (!msa9_switch)
 		exit(TEST_SKIP);
 
@@ -2650,6 +3073,7 @@ int main(void)
 	ecdsa_test();
 
 	return TEST_SUCC;
+#endif
 }
 
 #endif
