@@ -19,6 +19,7 @@
 #include <stdint.h>
 #include <openssl/crypto.h>
 #include <openssl/rsa.h>
+#include <openssl/evp.h>
 
 #include <openssl/opensslconf.h>
 #ifdef OPENSSL_FIPS
@@ -28,6 +29,13 @@
 #include "fips.h"
 #include "s390_rsa.h"
 #include "s390_prng.h"
+#include "s390_crypto.h"
+
+#if OPENSSL_VERSION_PREREQ(3, 0)
+#include <openssl/core_names.h>
+#include <openssl/param_build.h>
+extern OSSL_LIB_CTX *openssl_libctx;
+#endif
 
 #if defined(NO_SW_FALLBACKS)
 #define UNUSED(var)			((void)(var))
@@ -54,11 +62,12 @@ static unsigned int mod_expo_sw(int arg_length, char *arg, int exp_length,
 				int *res_length, char *res, BN_CTX *ctx);
 #endif /* NO_SW_FALLBACKS */
 
+#if !OPENSSL_VERSION_PREREQ(3, 0)
 RSA* rsa_key_generate(unsigned int modulus_bit_length,
 		      unsigned long *public_exponent)
 {
 #ifdef ICA_FIPS
-	if ((fips & ICA_FIPS_MODE) && (!FIPS_mode()))
+	if ((fips & ICA_FIPS_MODE) && (!openssl_in_fips_mode()))
 		return NULL;
 #endif /* ICA_FIPS */
 
@@ -78,17 +87,12 @@ RSA* rsa_key_generate(unsigned int modulus_bit_length,
 		BN_free(exp);
 		return NULL;
 	}
-#if OPENSSL_VERSION_NUMBER >= 0x10100000L
 	BN_GENCB *cb = BN_GENCB_new();
 	if(!cb) {
 		RSA_free(rsa);
 		BN_free(exp);
 		return NULL;
 	}
-#else
-	BN_GENCB dummy;
-	BN_GENCB *cb = &dummy;
-#endif /* OPENSSL_VERSION_NUMBER */
 
 	BN_set_word(exp, *public_exponent);
 	BN_GENCB_set_old(cb, NULL, NULL);
@@ -99,11 +103,62 @@ RSA* rsa_key_generate(unsigned int modulus_bit_length,
 	}
 
 	BN_free(exp);
-#if OPENSSL_VERSION_NUMBER >= 0x10100000L
 	BN_GENCB_free(cb);
-#endif /* OPENSSL_VERSION_NUMBER */
 	return rsa;
 }
+#else /* openssl 3.0 */
+EVP_PKEY* rsa_key_generate(unsigned int modulus_bit_length,
+		      unsigned long *public_exponent)
+{
+	EVP_PKEY *pkey = NULL;
+	EVP_PKEY_CTX *pctx = NULL;
+	BIGNUM *e = NULL;
+
+#ifdef ICA_FIPS
+	if ((fips & ICA_FIPS_MODE) && (!openssl_in_fips_mode()))
+		return NULL;
+#endif /* ICA_FIPS */
+
+	if (*public_exponent == 0)
+	{
+		do {
+			if (s390_prng((unsigned char*)public_exponent,
+			    sizeof(unsigned long)) != 0)
+				return NULL;
+		} while (*public_exponent <= 2 || !(*public_exponent % 2));
+	}
+
+	e = BN_bin2bn(public_exponent, sizeof(unsigned long), NULL);
+	if (e == NULL) {
+		goto done;
+	}
+
+	pctx = EVP_PKEY_CTX_new_id(EVP_PKEY_RSA, NULL);
+	if (pctx == NULL) {
+		goto done;
+	}
+
+	if (EVP_PKEY_keygen_init(pctx) != 1 ||
+		EVP_PKEY_CTX_set_rsa_keygen_bits(pctx, modulus_bit_length) != 1 ||
+		EVP_PKEY_CTX_set1_rsa_keygen_pubexp(pctx, e) != 1) {
+		goto done;
+	}
+
+	if (EVP_PKEY_keygen(pctx, &pkey) != 1) {
+		if (pkey)
+			EVP_PKEY_free(pkey);
+		pkey = NULL;
+	}
+
+done:
+	if (pctx != NULL)
+		EVP_PKEY_CTX_free(pctx);
+	if (e != NULL)
+		BN_free(e);
+
+	return pkey;
+}
+#endif
 
 /**
  * @brief Create a RSA modulus/expo key pair
@@ -128,26 +183,50 @@ unsigned int rsa_key_generate_mod_expo(ica_adapter_handle_t deviceHandle,
 				       ica_rsa_key_mod_expo_t *private_key)
 {
 	(void)deviceHandle;	/* suppress unused param warning */
+	int rc = 0;
 
 #ifdef ICA_FIPS
-	if ((fips & ICA_FIPS_MODE) && (!FIPS_mode()))
+	if ((fips & ICA_FIPS_MODE) && (!openssl_in_fips_mode()))
 		return EACCES;
+#if OPENSSL_VERSION_PREREQ(3, 0)
+	if ((fips & ICA_FIPS_MODE) && (modulus_bit_length <= 1024))
+		return EPERM;
+#endif
 #endif /* ICA_FIPS */
 
+	BEGIN_OPENSSL_LIBCTX(openssl_libctx, rc);
+
+#if !OPENSSL_VERSION_PREREQ(3, 0)
+	const BIGNUM *n, *d;
+
 	RSA *rsa = rsa_key_generate(modulus_bit_length,
 				    (unsigned long*)(public_key->exponent +
 				    public_key->key_length -
 				    sizeof(unsigned long)));
-	if (!rsa)
-		return errno;
+	if (!rsa) {
+		rc = EFAULT;
+		goto err;
+	}
 
-	const BIGNUM *n, *d;
-#if OPENSSL_VERSION_NUMBER >= 0x10100000L
 	RSA_get0_key(rsa, &n, NULL,  &d);
 #else
-	n = rsa->n;
-	d = rsa->d;
-#endif /* OPENSSL_VERSION_NUMBER */
+	BIGNUM *n = NULL, *d = NULL;
+
+	EVP_PKEY *pkey = rsa_key_generate(modulus_bit_length,
+				    (unsigned long*)(public_key->exponent +
+				    public_key->key_length -
+				    sizeof(unsigned long)));
+	if (!pkey) {
+		rc = EFAULT;
+		goto err;
+	}
+
+	if (!EVP_PKEY_get_bn_param(pkey, OSSL_PKEY_PARAM_RSA_N, &n) ||
+		!EVP_PKEY_get_bn_param(pkey, OSSL_PKEY_PARAM_RSA_D, &d)) {
+		rc = EFAULT;
+		goto err;
+	}
+#endif
 
 	/* Set key buffers zero to make sure there is no
 	 * unneeded junk in between.
@@ -175,9 +254,16 @@ unsigned int rsa_key_generate_mod_expo(ica_adapter_handle_t deviceHandle,
 		offset = 0;
 	BN_bn2bin(d, private_key->exponent + offset);
 
-	RSA_free(rsa);
+err:
 
-	return 0;
+#if !OPENSSL_VERSION_PREREQ(3, 0)
+	RSA_free(rsa);
+#else
+	EVP_PKEY_free(pkey);
+#endif
+
+	END_OPENSSL_LIBCTX(rc);
+	return rc;
 }
 
 /**
@@ -200,32 +286,56 @@ unsigned int rsa_key_generate_crt(ica_adapter_handle_t deviceHandle,
 				  ica_rsa_key_crt_t *private_key)
 {
 	(void)deviceHandle;	/* suppress unused param warning */
+	int rc = 0;
 
 #ifdef ICA_FIPS
-	if ((fips & ICA_FIPS_MODE) && (!FIPS_mode()))
+	if ((fips & ICA_FIPS_MODE) && (!openssl_in_fips_mode()))
 		return EACCES;
+#if OPENSSL_VERSION_PREREQ(3, 0)
+	if ((fips & ICA_FIPS_MODE) && (modulus_bit_length <= 1024))
+		return EPERM;
+#endif
 #endif /* ICA_FIPS */
 
+	BEGIN_OPENSSL_LIBCTX(openssl_libctx, rc);
+
+#if !OPENSSL_VERSION_PREREQ(3, 0)
+	const BIGNUM *n, *p, *q, *dmp1, *dmq1, *iqmp;
+
 	RSA *rsa = rsa_key_generate(modulus_bit_length,
 				    (unsigned long*)(public_key->exponent +
 				    public_key->key_length -
 				    sizeof(unsigned long)));
-	if (!rsa)
-		return errno;
+	if (!rsa) {
+		rc = EFAULT;
+		goto err;
+	}
 
-	const BIGNUM *n, *p, *q, *dmp1, *dmq1, *iqmp;
-#if OPENSSL_VERSION_NUMBER >= 0x10100000L
 	RSA_get0_key(rsa, &n, NULL,  NULL);
 	RSA_get0_factors(rsa, &p, &q);
 	RSA_get0_crt_params(rsa, &dmp1, &dmq1, &iqmp);
 #else
-	n = rsa->n;
-	p = rsa->p;
-	q = rsa->q;
-	dmp1 = rsa->dmp1;
-	dmq1 = rsa->dmq1;
-	iqmp = rsa->iqmp;
-#endif /* OPENSSL_VERSION_NUMBER */
+	BIGNUM *n = NULL, *p = NULL, *q = NULL, *dmp1 = NULL, *dmq1 = NULL, *iqmp = NULL;
+
+	EVP_PKEY *pkey = rsa_key_generate(modulus_bit_length,
+				    (unsigned long*)(public_key->exponent +
+				    public_key->key_length -
+				    sizeof(unsigned long)));
+	if (!pkey) {
+		rc = EFAULT;
+		goto err;
+	}
+
+	if (!EVP_PKEY_get_bn_param(pkey, OSSL_PKEY_PARAM_RSA_N, &n) ||
+		!EVP_PKEY_get_bn_param(pkey, OSSL_PKEY_PARAM_RSA_FACTOR1, &p) ||
+		!EVP_PKEY_get_bn_param(pkey, OSSL_PKEY_PARAM_RSA_FACTOR2, &q) ||
+		!EVP_PKEY_get_bn_param(pkey, OSSL_PKEY_PARAM_RSA_EXPONENT1, &dmp1) ||
+		!EVP_PKEY_get_bn_param(pkey, OSSL_PKEY_PARAM_RSA_EXPONENT2, &dmq1) ||
+		!EVP_PKEY_get_bn_param(pkey, OSSL_PKEY_PARAM_RSA_COEFFICIENT1, &iqmp)) {
+		rc = EFAULT;
+		goto err;
+	}
+#endif
 
 	/* Public exponent has already been set, no need to do this here.
 	 * For public key, only modulus needs to be set.
@@ -296,9 +406,15 @@ unsigned int rsa_key_generate_crt(ica_adapter_handle_t deviceHandle,
 
 	BN_bn2bin(iqmp, private_key->qInverse + 8 + offset);
 
+err:
+#if !OPENSSL_VERSION_PREREQ(3, 0)
 	RSA_free(rsa);
+#else
+	EVP_PKEY_free(pkey);
+#endif
 
-	return 0;
+	END_OPENSSL_LIBCTX(rc);
+	return rc;
 }
 
 #ifndef NO_SW_FALLBACKS
@@ -311,12 +427,12 @@ unsigned int rsa_mod_mult_sw(ica_rsa_modmult_t *pMul)
 	BN_CTX *ctx = NULL;
 
 #ifdef ICA_FIPS
-	if ((fips & ICA_FIPS_MODE) && (!FIPS_mode()))
+	if ((fips & ICA_FIPS_MODE) && (!openssl_in_fips_mode()))
 		return EACCES;
 #endif /* ICA_FIPS */
 
 	if ((ctx = BN_CTX_new()) == NULL) {
-		return errno;
+		return EFAULT;
 	}
 
 	rc = mod_mul_sw(pMul->inputdatalength, pMul->inputdata,
@@ -347,7 +463,7 @@ static unsigned int mod_mul_sw(int fc_1_length, char *fc1, int fc_2_length,
 	BIGNUM *b_res = NULL;
 
 #ifdef ICA_FIPS
-	if ((fips & ICA_FIPS_MODE) && (!FIPS_mode()))
+	if ((fips & ICA_FIPS_MODE) && (!openssl_in_fips_mode()))
 		return EACCES;
 #endif /* ICA_FIPS */
 
@@ -420,12 +536,12 @@ unsigned int rsa_mod_expo_sw(ica_rsa_modexpo_t *pMex)
 	BN_CTX *ctx = NULL;
 
 #ifdef ICA_FIPS
-	if ((fips & ICA_FIPS_MODE) && (!FIPS_mode()))
+	if ((fips & ICA_FIPS_MODE) && (!openssl_in_fips_mode()))
 		return EACCES;
 #endif /* ICA_FIPS */
 
 	if ((ctx = BN_CTX_new()) == NULL) {
-		return errno;
+		return EFAULT;
 	}
 
 	/* check if modulus value > data value */
@@ -486,7 +602,7 @@ static unsigned int mod_expo_sw(int arg_length, char *arg, int exp_length,
 	int mod_expo_rc = 1;
 
 #ifdef ICA_FIPS
-	if ((fips & ICA_FIPS_MODE) && (!FIPS_mode()))
+	if ((fips & ICA_FIPS_MODE) && (!openssl_in_fips_mode()))
 		return EACCES;
 #endif /* ICA_FIPS */
 
@@ -573,7 +689,7 @@ unsigned int rsa_crt_sw(ica_rsa_modexpo_crt_t * pCrt)
 	BN_CTX *ctx = NULL;
 
 #ifdef ICA_FIPS
-	if ((fips & ICA_FIPS_MODE) && (!FIPS_mode()))
+	if ((fips & ICA_FIPS_MODE) && (!openssl_in_fips_mode()))
 		return EACCES;
 #endif /* ICA_FIPS */
 
@@ -592,7 +708,7 @@ unsigned int rsa_crt_sw(ica_rsa_modexpo_crt_t * pCrt)
 	int temp_length = sizeof(temp);
 
 	if ((ctx = BN_CTX_new()) == NULL) {
-		return errno;
+		return EFAULT;
 	}
 
 	memset(ir1, 0, sizeof(ir1));
@@ -720,7 +836,7 @@ static unsigned int mod_sw(int arg_length, char *arg, int mod_length,
 	BIGNUM *b_res = NULL;
 
 #ifdef ICA_FIPS
-	if ((fips & ICA_FIPS_MODE) && (!FIPS_mode()))
+	if ((fips & ICA_FIPS_MODE) && (!openssl_in_fips_mode()))
 		return EACCES;
 #endif /* ICA_FIPS */
 
@@ -801,7 +917,7 @@ static unsigned int mod_sub_sw(int min_length, char *minu, int sub_length,
 	int min_size, sub_size, dif_size;
 
 #ifdef ICA_FIPS
-	if ((fips & ICA_FIPS_MODE) && (!FIPS_mode()))
+	if ((fips & ICA_FIPS_MODE) && (!openssl_in_fips_mode()))
 		return EACCES;
 #endif /* ICA_FIPS */
 
@@ -907,7 +1023,7 @@ static unsigned int add_sw(int aug_length, char *aug, int add_length,
 	BN_CTX_start(ctx);
 
 #ifdef ICA_FIPS
-	if ((fips & ICA_FIPS_MODE) && (!FIPS_mode()))
+	if ((fips & ICA_FIPS_MODE) && (!openssl_in_fips_mode()))
 		return EACCES;
 #endif /* ICA_FIPS */
 
@@ -976,7 +1092,7 @@ static unsigned int mul_sw(int fc_1_length, char *fc1, int fc_2_length,
 	BN_CTX_start(ctx);
 
 #ifdef ICA_FIPS
-	if ((fips & ICA_FIPS_MODE) && (!FIPS_mode()))
+	if ((fips & ICA_FIPS_MODE) && (!openssl_in_fips_mode()))
 		return EACCES;
 #endif /* ICA_FIPS */
 
